/*
 * The Npic library
 *
 * Copyright (C) 2003 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software under the terms of the 
 * GNU Lesser General Public License (LGPL) version 2.1.
*/

/* DO NOT EDIT !!!  Generated by npic-templa from "dist_skel_saito.ct" */

/*
 * dist_skel_saito.c - 10/02/2009
 *
 * Saito and Toriwaki's Euclidean Skeleton
 *
 * This skeleton Sk is defined in [1]. We compute Sk by using optimal SERDT 
 * from [2], following the description in [3]. Basically, Sk is the set of
 * elliptic paraboloids that belong to the upper enveloppe of all of the
 * elliptic paraboloids whose heights are given by SEDT.
 * The Sk can be used to compute RDMA (reduced discrete medial axis) by a 
 * filtering process, see dist_rdma_coeurjo.c
 *
 * References:
 *
 * [1] T. Saito et J.I. Toriwaki, 1994. "Reverse Distance Transformation and
 *     Skeletons based upon the Euclidean Metric for n-Dimensionnal Digital
 *     Pictures". IECE Trans. Information & Systems, 77(9):1005-1016.
 *
 * [2] T. Hirata, 1996. "A unified linear-time algorithm for computing  
 *     distance maps". Information Processing Letters, 58:129-133.
 *
 * [3] D. Coeurjolly and A. Montanvert, 2007. "Optimal Separable Algorithms to 
 *     Compute the Reverse Euclidean Distance Transformation and Discrete Medial
 *     Axis in Arbitrary Dimension". IEEE Trans. on PAMI, 29(3):437-448.
*/


#include <npic.h>


/*--------------------- P U B L I C - I N T E R F A C E ----------------------*/

/*
 * Compute the Saito and Toriwaki's Euclidean Skeleton Sk from SEDT.
 * Input:  np is a SEDT, computed with NpicSEDT_Hirata from a binary shape,
 * Output: np is Sk
 *
 * We pick up NpicSEDT_Hirata_rev,
 *  - using Npic_Hirata_scan_rev for each dimension but the last,
 *  - using Npic_Hirata_scan_skel for the last.
 *
 * Do nothing if np is not ok. Set ok on error. Verbose.
 * Return computation time in second.microsecond
*/

double NpicSkel_Saito (Npic_image *np)
{
    double time1, time2;
    int L;

    if (NpicImageIsOK (np, __func__) != NPIC_SUCCESS) return 0;

    L = np->gen.xmax;
    if (np->gen.ymax > L) L = np->gen.ymax;
    if (np->gen.dim >= 3 && np->gen.zmax > L) L = np->gen.zmax;
    if (np->gen.dim >= 4 && np->gen.tmax > L) L = np->gen.tmax;
    if (np->gen.dim >= 5 && np->gen.smax > L) L = np->gen.smax;
    if (np->gen.dim >= 6 && np->gen.rmax > L) L = np->gen.rmax;

    time1 = NpicGetTime();

    switch (np->type) {

        case NPIC_IMAGE_2L : {  /* - - - - - - - - - - - - - - - - - */
        
            Npic_image_2l *p = NpicCastImage(np);
            Npic_l Cv[L];  /* values */
            int y, x;

#           ifdef NPIC_USE_OMP
#           pragma omp parallel private(y, x, Cv)
            {
#           endif

            /*
             * First scan in x
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (y = 0; y < p->ymax; y++)
            {
                for (x = 0; x < p->xmax; x++)
                    Cv[x] = p->pix[y][x];

                Npic_Hirata_scan_rev (Cv, p->xmax);

                for (x = 0; x < p->xmax; x++)
                    p->pix[y][x] = Cv[x];
            }

            /*
             * Scan in y
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (x = 0; x < p->xmax; x++)
            {
                for (y = 0; y < p->ymax; y++)
                    Cv[y] = p->pix[y][x];

                Npic_Hirata_scan_skel (Cv, p->ymax);

                for (y = 0; y < p->ymax; y++)
                    p->pix[y][x] = Cv[y];
            }

#           ifdef NPIC_USE_OMP
            }  /* omp parallel */
#           endif
            break;
        }

        case NPIC_IMAGE_3L : {  /* - - - - - - - - - - - - - - - - - */
        
            Npic_image_3l *p = NpicCastImage(np);
            Npic_l Cv[L];  /* values */
            int z, y, x;

#           ifdef NPIC_USE_OMP
#           pragma omp parallel private(z, y, x, Cv)
            {
#           endif

            /*
             * First scan in x
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (z = 0; z < p->zmax; z++)
            for (y = 0; y < p->ymax; y++)
            {
                for (x = 0; x < p->xmax; x++)
                    Cv[x] = p->pix[z][y][x];

                Npic_Hirata_scan_rev (Cv, p->xmax);

                for (x = 0; x < p->xmax; x++)
                    p->pix[z][y][x] = Cv[x];
            }

            /*
             * Scan in y
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (z = 0; z < p->zmax; z++)
            for (x = 0; x < p->xmax; x++)
            {
                for (y = 0; y < p->ymax; y++)
                    Cv[y] = p->pix[z][y][x];

                Npic_Hirata_scan_rev (Cv, p->ymax);

                for (y = 0; y < p->ymax; y++)
                    p->pix[z][y][x] = Cv[y];
            }

            /*
             * Scan in z
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (y = 0; y < p->ymax; y++)
            for (x = 0; x < p->xmax; x++)
            {
                for (z = 0; z < p->zmax; z++)
                    Cv[z] = p->pix[z][y][x];

                Npic_Hirata_scan_skel (Cv, p->zmax);

                for (z = 0; z < p->zmax; z++)
                    p->pix[z][y][x] = Cv[z];
            }

#           ifdef NPIC_USE_OMP
            }  /* omp parallel */
#           endif
            break;
        }

        case NPIC_IMAGE_4L : {  /* - - - - - - - - - - - - - - - - - */
        
            Npic_image_4l *p = NpicCastImage(np);
            Npic_l Cv[L];  /* values */
            int t, z, y, x;

#           ifdef NPIC_USE_OMP
#           pragma omp parallel private(t, z, y, x, Cv)
            {
#           endif

            /*
             * First scan in x
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (t = 0; t < p->tmax; t++)
            for (z = 0; z < p->zmax; z++)
            for (y = 0; y < p->ymax; y++)
            {
                for (x = 0; x < p->xmax; x++)
                    Cv[x] = p->pix[t][z][y][x];

                Npic_Hirata_scan_rev (Cv, p->xmax);

                for (x = 0; x < p->xmax; x++)
                    p->pix[t][z][y][x] = Cv[x];
            }

            /*
             * Scan in y
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (t = 0; t < p->tmax; t++)
            for (z = 0; z < p->zmax; z++)
            for (x = 0; x < p->xmax; x++)
            {
                for (y = 0; y < p->ymax; y++)
                    Cv[y] = p->pix[t][z][y][x];

                Npic_Hirata_scan_rev (Cv, p->ymax);

                for (y = 0; y < p->ymax; y++)
                    p->pix[t][z][y][x] = Cv[y];
            }

            /*
             * Scan in z
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (t = 0; t < p->tmax; t++)
            for (y = 0; y < p->ymax; y++)
            for (x = 0; x < p->xmax; x++)
            {
                for (z = 0; z < p->zmax; z++)
                    Cv[z] = p->pix[t][z][y][x];

                Npic_Hirata_scan_rev (Cv, p->zmax);

                for (z = 0; z < p->zmax; z++)
                    p->pix[t][z][y][x] = Cv[z];
            }

            /*
             * Scan in t
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (z = 0; z < p->zmax; z++)
            for (y = 0; y < p->ymax; y++)
            for (x = 0; x < p->xmax; x++)
            {
                for (t = 0; t < p->tmax; t++)
                    Cv[t] = p->pix[t][z][y][x];

                Npic_Hirata_scan_skel (Cv, p->tmax);

                for (t = 0; t < p->tmax; t++)
                    p->pix[t][z][y][x] = Cv[t];
            }

#           ifdef NPIC_USE_OMP
            }  /* omp parallel */
#           endif
            break;
        }

        case NPIC_IMAGE_5L : {  /* - - - - - - - - - - - - - - - - - */
        
            Npic_image_5l *p = NpicCastImage(np);
            Npic_l Cv[L];  /* values */
            int s, t, z, y, x;

#           ifdef NPIC_USE_OMP
#           pragma omp parallel private(s, t, z, y, x, Cv)
            {
#           endif

            /*
             * First scan in x
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (s = 0; s < p->smax; s++)
            for (t = 0; t < p->tmax; t++)
            for (z = 0; z < p->zmax; z++)
            for (y = 0; y < p->ymax; y++)
            {
                for (x = 0; x < p->xmax; x++)
                    Cv[x] = p->pix[s][t][z][y][x];

                Npic_Hirata_scan_rev (Cv, p->xmax);

                for (x = 0; x < p->xmax; x++)
                    p->pix[s][t][z][y][x] = Cv[x];
            }

            /*
             * Scan in y
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (s = 0; s < p->smax; s++)
            for (t = 0; t < p->tmax; t++)
            for (z = 0; z < p->zmax; z++)
            for (x = 0; x < p->xmax; x++)
            {
                for (y = 0; y < p->ymax; y++)
                    Cv[y] = p->pix[s][t][z][y][x];

                Npic_Hirata_scan_rev (Cv, p->ymax);

                for (y = 0; y < p->ymax; y++)
                    p->pix[s][t][z][y][x] = Cv[y];
            }

            /*
             * Scan in z
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (s = 0; s < p->smax; s++)
            for (t = 0; t < p->tmax; t++)
            for (y = 0; y < p->ymax; y++)
            for (x = 0; x < p->xmax; x++)
            {
                for (z = 0; z < p->zmax; z++)
                    Cv[z] = p->pix[s][t][z][y][x];

                Npic_Hirata_scan_rev (Cv, p->zmax);

                for (z = 0; z < p->zmax; z++)
                    p->pix[s][t][z][y][x] = Cv[z];
            }

            /*
             * Scan in t
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (s = 0; s < p->smax; s++)
            for (z = 0; z < p->zmax; z++)
            for (y = 0; y < p->ymax; y++)
            for (x = 0; x < p->xmax; x++)
            {
                for (t = 0; t < p->tmax; t++)
                    Cv[t] = p->pix[s][t][z][y][x];

                Npic_Hirata_scan_rev (Cv, p->tmax);

                for (t = 0; t < p->tmax; t++)
                    p->pix[s][t][z][y][x] = Cv[t];
            }

            /*
             * Scan in s
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (t = 0; t < p->tmax; t++)
            for (z = 0; z < p->zmax; z++)
            for (y = 0; y < p->ymax; y++)
            for (x = 0; x < p->xmax; x++)
            {
                for (s = 0; s < p->smax; s++)
                    Cv[s] = p->pix[s][t][z][y][x];

                Npic_Hirata_scan_skel (Cv, p->smax);

                for (s = 0; s < p->smax; s++)
                    p->pix[s][t][z][y][x] = Cv[s];
            }

#           ifdef NPIC_USE_OMP
            }  /* omp parallel */
#           endif
            break;
        }

        case NPIC_IMAGE_6L : {  /* - - - - - - - - - - - - - - - - - */
        
            Npic_image_6l *p = NpicCastImage(np);
            Npic_l Cv[L];  /* values */
            int r, s, t, z, y, x;

#           ifdef NPIC_USE_OMP
#           pragma omp parallel private(r, s, t, z, y, x, Cv)
            {
#           endif

            /*
             * First scan in x
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (r = 0; r < p->rmax; r++)
            for (s = 0; s < p->smax; s++)
            for (t = 0; t < p->tmax; t++)
            for (z = 0; z < p->zmax; z++)
            for (y = 0; y < p->ymax; y++)
            {
                for (x = 0; x < p->xmax; x++)
                    Cv[x] = p->pix[r][s][t][z][y][x];

                Npic_Hirata_scan_rev (Cv, p->xmax);

                for (x = 0; x < p->xmax; x++)
                    p->pix[r][s][t][z][y][x] = Cv[x];
            }

            /*
             * Scan in y
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (r = 0; r < p->rmax; r++)
            for (s = 0; s < p->smax; s++)
            for (t = 0; t < p->tmax; t++)
            for (z = 0; z < p->zmax; z++)
            for (x = 0; x < p->xmax; x++)
            {
                for (y = 0; y < p->ymax; y++)
                    Cv[y] = p->pix[r][s][t][z][y][x];

                Npic_Hirata_scan_rev (Cv, p->ymax);

                for (y = 0; y < p->ymax; y++)
                    p->pix[r][s][t][z][y][x] = Cv[y];
            }

            /*
             * Scan in z
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (r = 0; r < p->rmax; r++)
            for (s = 0; s < p->smax; s++)
            for (t = 0; t < p->tmax; t++)
            for (y = 0; y < p->ymax; y++)
            for (x = 0; x < p->xmax; x++)
            {
                for (z = 0; z < p->zmax; z++)
                    Cv[z] = p->pix[r][s][t][z][y][x];

                Npic_Hirata_scan_rev (Cv, p->zmax);

                for (z = 0; z < p->zmax; z++)
                    p->pix[r][s][t][z][y][x] = Cv[z];
            }

            /*
             * Scan in t
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (r = 0; r < p->rmax; r++)
            for (s = 0; s < p->smax; s++)
            for (z = 0; z < p->zmax; z++)
            for (y = 0; y < p->ymax; y++)
            for (x = 0; x < p->xmax; x++)
            {
                for (t = 0; t < p->tmax; t++)
                    Cv[t] = p->pix[r][s][t][z][y][x];

                Npic_Hirata_scan_rev (Cv, p->tmax);

                for (t = 0; t < p->tmax; t++)
                    p->pix[r][s][t][z][y][x] = Cv[t];
            }

            /*
             * Scan in s
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (r = 0; r < p->rmax; r++)
            for (t = 0; t < p->tmax; t++)
            for (z = 0; z < p->zmax; z++)
            for (y = 0; y < p->ymax; y++)
            for (x = 0; x < p->xmax; x++)
            {
                for (s = 0; s < p->smax; s++)
                    Cv[s] = p->pix[r][s][t][z][y][x];

                Npic_Hirata_scan_rev (Cv, p->smax);

                for (s = 0; s < p->smax; s++)
                    p->pix[r][s][t][z][y][x] = Cv[s];
            }

            /*
             * Scan in r
            */
#           ifdef NPIC_USE_OMP
#           pragma omp for schedule(dynamic)
#           endif
            for (s = 0; s < p->smax; s++)
            for (t = 0; t < p->tmax; t++)
            for (z = 0; z < p->zmax; z++)
            for (y = 0; y < p->ymax; y++)
            for (x = 0; x < p->xmax; x++)
            {
                for (r = 0; r < p->rmax; r++)
                    Cv[r] = p->pix[r][s][t][z][y][x];

                Npic_Hirata_scan_skel (Cv, p->rmax);

                for (r = 0; r < p->rmax; r++)
                    p->pix[r][s][t][z][y][x] = Cv[r];
            }

#           ifdef NPIC_USE_OMP
            }  /* omp parallel */
#           endif
            break;
        }

        default :
            np->gen.ok = NpicError (__func__, NPIC_ERR_UNEX_NPIC, "");
    }

    time2 = NpicGetTime();
    return time2-time1;
}



/*-------------------- P R I V A T E - F U N C T I O N S ---------------------*/

/*
 * Scan column Cv[0..cmax-1] and keep Sk points.
 * Used in NpicSkel_Saito.
*/

void Npic_Hirata_scan_skel (Npic_l *Cv, int cmax)
{
    Npic_l k, Sv[cmax];     /* values */
    int j, Si[cmax], Sn;    /* index, number */
    double Sr[cmax];        /* intersections */

    Sn = 0;  /* Empty stacks */

    /* Compute stacks indices Si[Sn] and RevSEDT values Sv[Sn] */
    for (j = 0; j < cmax; j++)
    {
        k = Cv[j]; 

        while (Sn >= 2 && Npic_Hirata_inter (Si[Sn-1], -Sv[Sn-1], j, -k) < Sr[Sn-1])
            Sn--;  /* pop */

        /* push */
        Si[Sn] = j; Sv[Sn] = k;
        if (Sn >= 1) 
            Sr[Sn] = Npic_Hirata_inter (Si[Sn-1], -Sv[Sn-1], Si[Sn], -Sv[Sn]);
        Sn++;
    }

    /* Empty Cv[] */
    for (j = 0; j < cmax; j++) Cv[j] = 0;

    /* Keep Sk points in Cv[] */
    for (j = 0; j < Sn; j++) Cv[ Si[j] ] = Sv[j];
}


/*----------------------------------------------------------------------------*/

